{% if issues %}
{% for issue in issues %}
==== {{ issue.title }} ====
SWC ID: {{ issue['swc-id'] }}
Severity: {{ issue.severity }}
Contract: {{ issue.contract | default("Unknown") }}
Function name: {{ issue.function }}
PC address: {{ issue.address }}
Estimated Gas Usage: {{ issue.min_gas_used }} - {{ issue.max_gas_used }}
{{ issue.description }}
--------------------
{% if issue.filename and issue.lineno %}
In file: {{ issue.filename }}:{{ issue.lineno }}
{% endif %}
{% if issue.code %}

{{ issue.code }}

--------------------
{% endif %}
{% if verbose and issue.debug %}
--------------------
Transaction Sequence:

{{ issue.debug }}
{% endif %}

{% endfor %}
{% else %}
The analysis was completed successfully. No issues were detected.
{% endif %}
